import sys,os
fpop_path = os.path.abspath(os.path.join(os.path.dirname(__file__),".."))
sys.path.insert(0,fpop_path)
import fpop
upload_python_packages = [os.path.join(fpop_path, 'fpop')]
if os.getenv('SKIP_UT_WITH_DFLOW'):
    if(os.getenv('SKIP_UT_WITH_DFLOW')=='0'):
        skip_ut_with_dflow=0
    else:
        skip_ut_with_dflow=1
    skip_ut_with_dflow_reason = 'skip because environment variable SKIP_UT_WITH_DFLOW is set to non-zero'
else:
    skip_ut_with_dflow = False
    skip_ut_with_dflow_reason = ''
# one needs to set proper values for the following variable.
default_image = 'dptechnology/dpgen2:latest'
if os.getenv("DFLOW_DEBUG"):
    from dflow import (
        config,
    )
    config["mode"] = "debug"



from fpop.prep_fp import PrepFp
from fpop.run_fp import RunFp
import dpdata, sys, subprocess, os, shutil
from pathlib import Path
from dflow.utils import run_command
from typing import (
    Any,
    Tuple,
    List,
    Set,
    Dict,
    Optional,
    Union,
)
import numpy as np
from dargs import (
    dargs, 
    Argument, 
    Variant, 
    ArgumentEncoder,
)
from dflow.python import (
    OP,
    OPIO,
    OPIOSign,
    Artifact,
    TransientError,
    FatalError,
    BigParameter,
)

class VaspInputs():
    def __init__(
            self,
            kspacing : Union[float, List[float]],
            incar : str,
            pp_files : Dict[str, str],
            kgamma : bool = True,
    ):
        """
        Parameters
        ----------
        kspacing : Union[float, List[float]]
                The kspacing. If it is a number, then three directions use the same
                ksapcing, otherwise it is a list of three numbers, specifying the
                kspacing used in the x, y and z dimension.
        incar: str
                A template INCAR file. 
        pp_files : Dict[str,str]
                The potcar files for the elements. For example
                { 
                   "H" : "/path/to/POTCAR_H",
                   "O" : "/path/to/POTCAR_O",
                }
        kgamma : bool
                K-mesh includes the gamma point
        """
        self.kspacing = kspacing
        self.kgamma = kgamma
        self.incar_from_file(incar)
        self.potcars_from_file(pp_files)

    @property
    def incar_template(self):
        return self._incar_template

    @property
    def potcars(self):
        return self._potcars

    def incar_from_file(
            self,
            fname : str,
    ):
        self._incar_template = Path(fname).read_text()

    def potcars_from_file(
            self,
            dict_fnames : Dict[str,str],
    ):
        self._potcars = {}
        for kk,vv in dict_fnames.items():
            self._potcars[kk] = Path(vv).read_text()            

    def make_potcar(
            self, 
            atom_names,
    ) -> str:        
        potcar_contents = []
        for nn in atom_names:
            potcar_contents.append(self._potcars[nn])
        return "".join(potcar_contents)            

    def make_kpoints(
            self,
            box : np.ndarray,
    ) -> str:
        return make_kspacing_kpoints(box, self.kspacing, self.kgamma)

    @staticmethod
    def args():
        doc_pp_files = 'The pseudopotential files set by a dict, e.g. {"Al" : "path/to/the/al/pp/file", "Mg" : "path/to/the/mg/pp/file"}'
        doc_incar = "The path to the template incar file"
        doc_kspacing = "The spacing of k-point sampling. `ksapcing` will overwrite the incar template"
        doc_kgamma = "If the k-mesh includes the gamma point. `kgamma` will overwrite the incar template"
        return [
            Argument("incar", str, optional=False, doc=doc_pp_files),
            Argument("pp_files", dict, optional=False, doc=doc_pp_files),
            Argument("kspacing", float, optional=False, doc=doc_kspacing),
            Argument("kgamma", bool, optional=True, default=True, doc=doc_kgamma),
        ]

def make_kspacing_kpoints(box, kspacing, kgamma) :
    if type(kspacing) is not list:
        kspacing = [kspacing, kspacing, kspacing]
    box = np.array(box)
    rbox = _reciprocal_box(box)
    kpoints = [max(1,(np.ceil(2 * np.pi * np.linalg.norm(ii) / ks).astype(int))) for ii,ks in zip(rbox,kspacing)]
    ret = _make_vasp_kpoints(kpoints, kgamma)
    return ret


def _make_vasp_kp_gamma(kpoints):
    ret = ""
    ret += "Automatic mesh\n"
    ret += "0\n"
    ret += "Gamma\n"
    ret += "%d %d %d\n" % (kpoints[0], kpoints[1], kpoints[2])
    ret += "0  0  0\n"
    return ret

def _make_vasp_kp_mp(kpoints):
    ret = ""
    ret += "K-Points\n"
    ret += "0\n"
    ret += "Monkhorst Pack\n"
    ret += "%d %d %d\n" % (kpoints[0], kpoints[1], kpoints[2])
    ret += "0  0  0\n"
    return ret

def _make_vasp_kpoints (kpoints, kgamma = False) :
    if kgamma :
        ret = _make_vasp_kp_gamma(kpoints)
    else :
        ret = _make_vasp_kp_mp(kpoints)
    return ret
    
def _reciprocal_box(box) :
    rbox = np.linalg.inv(box)
    rbox = rbox.T
    return rbox


class PrepVasp(PrepFp):
    def prep_task(
            self,
            conf_frame: dpdata.System,
            vasp_inputs: VaspInputs,
            prepare_image_config: Optional[Dict] = None,
            optional_input: Optional[Dict] = None,
            optional_artifact: Optional[Dict] = None,
    ):
        r"""Define how one Vasp task is prepared.

        Parameters
        ----------
        conf_frame : dpdata.System
            One frame of configuration in the dpdata format.
        inputs: VaspInputs
            The VaspInputs object handels all other input files of the task.
        prepare_image_config: Dict
            Definition of runtime parameters in the process of preparing tasks. 
        optional_input: 
            Other parameters the developers or users may need.For example:
            {
               "oonf_format": "vasp/poscar"
            }
            optional_input["conf_format"]: The format of the configurations which users give.  
        optional_artifact
            Other files that users or developers need.For example:
            {
               "ICONST": Path("./iconst")
            }
            In vasp part, all the files which are given in optional_artifact will be copied to the work directory. In this example, "INCAR","POTCAR","POSCAR","KPOINTS" and "ICONST" will be copied to the same directory. "./iconst" is the path where the target file exists.
        """

        conf_frame.to('vasp/poscar', 'POSCAR')
        Path('INCAR').write_text(
            vasp_inputs.incar_template
        )
        # fix the case when some element have 0 atom, e.g. H0O2
        tmp_frame = dpdata.System('POSCAR', fmt='vasp/poscar')
        Path('POTCAR').write_text(
            vasp_inputs.make_potcar(tmp_frame['atom_names'])
        )
        Path('KPOINTS').write_text(
            vasp_inputs.make_kpoints(conf_frame['cells'][0])
        )

        if optional_artifact:
            for file_name, file_path in optional_artifact.items():
                content = file_path.read_text()
                Path(file_name).write_text(content)


class RunVasp(RunFp):
    def input_files(self) -> List[str]:
        r'''The mandatory input files to run a vasp task.
        Returns
        -------
        files: List[str]
            A list of madatory input files names.
        '''
        return ["POSCAR", "INCAR", "POTCAR", "KPOINTS"]

    def run_task(
        self,
        backward_dir_name,
        log_name,
        backward_list: List[str],
        run_image_config: Optional[Dict]=None,
        optional_input: Optional[Dict]=None,
    ) -> str:
        r'''Defines how one FP task runs
        Parameters
        ----------
        backward_dir_name:
            The name of the directory which contains the backward files.
        log_name:
            The name of log file.
        backward_list:
            The output files the users need.For example: ["OUTCAR","CONTCAR"]
        run_image_config:
            Keyword args defined by the developer.For example:
            {
              "command": "source /opt/intel/oneapi/setvars.sh && mpirun -n 64 /opt/vasp.5.4.4/bin/vasp_std"
            }
        optional_input:
            The parameters developers need in runtime.For example:
            {
                "conf_format": "vasp/poscar"
            }
            optional_input["vasp/poscar"] is the format of the configurations that users give.
        
        Returns
        -------
        backward_dir_name: str
            The directory name which containers the files users need.
        '''
        if run_image_config:
            command = run_image_config["command"]
        else:
            command = "vasp_std"
        # run vasp
        command = " ".join([command, ">", log_name])
        ret, out, err = run_command(command, raise_error=False, try_bash=True,)
        if ret != 0:
            raise TransientError(
                "vasp failed\n", "out msg", out, "\n", "err msg", err, "\n"
            )
        if not self.check_run_success():
            raise TransientError(
                "vasp failed , we could not check the exact cause . Please check log file ."
            )
        os.makedirs(Path(backward_dir_name))
        shutil.copyfile(log_name,Path(backward_dir_name)/log_name)
        for ii in backward_list:
            shutil.copyfile(ii,Path(backward_dir_name)/ii)
        return backward_dir_name
    
    def check_run_success(self):
        with open("OUTCAR","r") as f:
            lines = f.readlines()
        if "Voluntary" in lines[-1]:
            return True
        else:
            return False



from dflow import (
    InputParameter,
    OutputParameter,
    Inputs,
    InputArtifact,
    Outputs,
    OutputArtifact,
    Workflow,
    Step,
    Steps,
    upload_artifact,
    download_artifact,
    argo_range,
    argo_len,
    argo_sequence,
)
from dflow.python import(
    PythonOPTemplate,
    OP,
    OPIO,
    OPIOSign,
    Artifact,
    Slices,
)

from dflow.plugins.dispatcher import DispatcherExecutor
import os,sys
from typing import Optional, Set, List
from pathlib import Path
from fpop.utils.step_config import (
    init_executor,
)

class PrepRunFp(Steps):
    def __init__(
        self,
        name : str,
        prep_op : OP,
        run_op : OP,
        prep_image : str,
        run_image : str,
        prep_template_config : Optional[dict] = None,
        prep_step_config : Optional[dict] = None,
        run_template_config : Optional[dict] = None,
        run_slice_config : Optional[dict] = None,
        run_step_config : Optional[dict] = None,
        upload_python_packages : Optional[List[str]] = None,
    ):
        self._input_parameters = {
            "inputs" : InputParameter(),
            "type_map" : InputParameter(),
            "backward_list" : InputParameter(),
            "prep_image_config" : InputParameter(type=dict , value={}),
            "run_image_config" : InputParameter(type=dict , value={}),
            "optional_input" : InputParameter(type=dict , value={}),
            "log_name" : InputParameter(type=str , value="log"),
            "backward_dir_name" : InputParameter(type=str , value="backward_dir"),
        }
        self._input_artifacts = {
            "confs" : InputArtifact(),
            "optional_artifact" : InputArtifact(optional=True), 
        }
        self._output_artifacts = {
            "backward_dirs" : OutputArtifact(),
        }

        super().__init__(
            name=name,
            inputs=Inputs(
                parameters=self._input_parameters,
                artifacts=self._input_artifacts,
            ),
            outputs=Outputs(
                artifacts=self._output_artifacts,
            )
        )

        self._keys = ['prep-fp','run-fp']
        self.step_keys = {'prep-fp':'prep-fp','run-fp':'run-fp-{{item}}'}

        self = _prep_run_fp(
            self,
            self.step_keys,
            prep_op,
            run_op,
            prep_image,
            run_image,
            prep_template_config,
            prep_step_config,
            run_template_config,
            run_slice_config,
            run_step_config,
            upload_python_packages = upload_python_packages,
        )

    @property
    def input_parameters(self):
        return self._input_parameters

    @property
    def input_artifacts(self):
        return self._input_artifacts

    @property
    def output_artifacts(self):
        return self._output_artifacts

    @property
    def keys(self):
        return self._keys

def _prep_run_fp(
        prep_run_steps,
        step_keys,
        prep_op : OP,
        run_op : OP,
        prep_image,
        run_image,
        prep_template_config : Optional[dict] = None,
        prep_step_config : Optional[dict] = None,
        run_template_config : Optional[dict] = None,
        run_slice_config : Optional[dict] = None,
        run_step_config : Optional[dict] = None,
        upload_python_packages : Optional[List[str]] = None,
):
    if not prep_template_config: prep_template_config = {}
    if not prep_step_config: prep_step_config = {}
    if not run_template_config: run_template_config = {}
    if not run_slice_config: run_slice_config = {}
    if not run_step_config: run_step_config = {}
    if "executor" in prep_step_config.keys():
        prep_executor = init_executor(prep_step_config.pop("executor"))
    else:
        prep_executor = None
    if "executor" in run_step_config.keys():
        run_executor = init_executor(run_step_config.pop("executor"))
    else:
        run_executor = None

    prep_fp = Step(
        'prep-fp' , 
        template=PythonOPTemplate(
            prep_op,
            output_artifact_archive={
                "task_paths": None
            },
            python_packages = upload_python_packages,
            image = prep_image,
            **prep_template_config,
        ),
        parameters={
            "prep_image_config" : prep_run_steps.inputs.parameters["prep_image_config"],
            "inputs" : prep_run_steps.inputs.parameters["inputs"],
            "type_map" : prep_run_steps.inputs.parameters["type_map"],
            "optional_input" : prep_run_steps.inputs.parameters["optional_input"],
        },
        artifacts={
            "confs" : prep_run_steps.inputs.artifacts['confs'],
            "optional_artifact" : prep_run_steps.inputs.artifacts['optional_artifact'],
        },
        key = step_keys['prep-fp'],
        executor = prep_executor,
        **prep_step_config,    
    )
    prep_run_steps.add(prep_fp)

    run_fp = Step(
        'run-fp',
        template=PythonOPTemplate(
            run_op,
            slices = Slices(
                "int('{{item}}')",
                input_parameter = ["task_name"],
                input_artifact = ["task_path"],
                output_artifact = ["backward_dir"],
                **run_slice_config,
            ),
            python_packages = upload_python_packages,
            image = run_image,
            **run_template_config,
        ),
        parameters={
            "run_image_config" : prep_run_steps.inputs.parameters["run_image_config"],
            "task_name" : prep_fp.outputs.parameters["task_names"],
            "backward_list" : prep_run_steps.inputs.parameters["backward_list"],
            "log_name" : prep_run_steps.inputs.parameters["log_name"],
            "backward_dir_name" : prep_run_steps.inputs.parameters["backward_dir_name"],
            "optional_input" : prep_run_steps.inputs.parameters["optional_input"],
        },
        artifacts={
            "task_path" : prep_fp.outputs.artifacts['task_paths'],
            "optional_artifact" : prep_run_steps.inputs.artifacts["optional_artifact"],
        },
        with_sequence=argo_sequence(argo_len(prep_fp.outputs.parameters["task_names"]), format='%06d'),
        key = step_keys['run-fp'],
        executor = run_executor,
        **run_step_config,
    )
    prep_run_steps.add(run_fp)

    prep_run_steps.outputs.artifacts["backward_dirs"]._from = run_fp.outputs.artifacts["backward_dir"]



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="fpop",
    version="0.0.5",
    author="Chengqian Zhang",
    author_email="2043899742@qq.com",
    description="Operators related to first-principles calculation",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/deepmodeling/fpop",
    packages=setuptools.find_packages(),
    install_requires=[
        "pydflow",
        "lbg",
        "dpdata",
        "numpy",
        "dargs",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
    provides=["fpop"],
    scripts=[]
)



import os
import numpy as np
import unittest

from dflow import (
    InputParameter,
    OutputParameter,
    Inputs,
    InputArtifact,
    Outputs,
    OutputArtifact,
    Workflow,
    Step,
    Steps,
    upload_artifact,
    download_artifact,
    S3Artifact,
    argo_range
)
from dflow.python import (
    PythonOPTemplate,
    OP,
    OPIO,
    OPIOSign,
    Artifact,
    upload_packages,
)

import time, shutil, dpdata
from pathlib import Path

from context import (
        fpop,
        default_image,
        upload_python_packages,
        skip_ut_with_dflow,
        skip_ut_with_dflow_reason,
        )
from fpop.vasp import PrepVasp,VaspInputs
from typing import List
from constants import POSCAR_1_content,POSCAR_2_content
upload_packages.append("../fpop")
upload_packages.append("./context.py")

def check_vasp_tasks(tcase, ntasks):
    cc = 0
    tdirs = []
    for ii in range (ntasks):
        tdir = "task.%06d" % cc
        tdirs.append(tdir)
        print(tdir)
        tcase.assertTrue(Path(tdir).is_dir())
        incar = Path(tdir)/'INCAR'
        poscar = Path(tdir)/'POSCAR'
        potcar = Path(tdir)/'POTCAR'
        kpoints = Path(tdir)/'KPOINTS'
        tcase.assertTrue(incar.is_file())
        tcase.assertTrue(poscar.is_file())
        tcase.assertTrue(potcar.is_file())
        tcase.assertTrue(kpoints.is_file())
        tcase.assertEqual(incar.read_text(),'here incar')
        tcase.assertEqual(potcar.read_text(),'here potcar')
        cc += 1
    return tdirs

def dump_conf_from_poscar(
        type, 
        conf_list
        ) -> List[str] :
    for ii in range(len(conf_list)):
        Path("POSCAR_%d"%ii).write_text(conf_list[ii])
    if type == "deepmd/npy":
        confs = []
        for ii in range(len(conf_list)):
            ls = dpdata.System("POSCAR_%d"%ii, fmt="vasp/poscar")
            ls.to_deepmd_npy("data.%03d"%ii)
            confs.append("data.%03d"%ii)
            os.remove("POSCAR_%d"%ii)
        return confs
    elif type == "vasp/poscar":
        confs = []
        for ii in range(len(conf_list)):
            confs.append("POSCAR_%d")
        return confs
    else:
        return []

class TestPrepVaspDpConf(unittest.TestCase):
    '''
    deepmd/npy format named ["data.000","data.001"].
    no optional_input or optional_artifact.
    '''
    def setUp(self):
        self.ntasks = 2
        confs = dump_conf_from_poscar("deepmd/npy",[POSCAR_1_content, POSCAR_2_content])
        self.confs = [Path(ii) for ii in confs]
        self.incar = 'incar'
        Path(self.incar).write_text('here incar')
        self.potcar = 'potcar'
        Path(self.potcar).write_text('here potcar')
        self.type_map = ['Na']
    
    def tearDown(self):
        for ii in range(self.ntasks):
            work_path = Path("task.%06d"%ii)
            if work_path.is_dir():
                shutil.rmtree(work_path)
        for ii in [self.incar,self.potcar]:
            if Path(ii).is_file:
                os.remove(ii)
        for ii in self.confs:
            if ii.is_dir():
                shutil.rmtree(ii)

    def test(self):
        op = PrepVasp()
        vasp_inputs = VaspInputs(
            0.3,
            self.incar,
            {'Na':self.potcar},
            True,
        )
        out = op.execute(
            OPIO(
                {
                    "prep_image_config" : {},
                    "optional_input" : {},
                    "confs" : self.confs,
                    "inputs" : vasp_inputs,
                    "type_map" : self.type_map,
                }
            )
        )
        tdirs = check_vasp_tasks(self, self.ntasks)
        self.assertEqual(tdirs, out['task_names'])
        self.assertEqual(tdirs, [str(ii) for ii in out['task_paths']])
     
    def testWithoutOptionalParameter(self):
        op = PrepVasp()
        vasp_inputs = VaspInputs(
            0.3,
            self.incar,
            {'Na':self.potcar},
            True,
        )
        out = op.execute(
            OPIO(
                {
                    "confs" : self.confs,
                    "inputs" : vasp_inputs,
                    "type_map" : self.type_map,
                }
            )
        )
        tdirs = check_vasp_tasks(self, self.ntasks)
        self.assertEqual(tdirs, out['task_names'])
        self.assertEqual(tdirs, [str(ii) for ii in out['task_paths']])
    
@unittest.skipIf(skip_ut_with_dflow, skip_ut_with_dflow_reason)
class TestPrepRunVaspPoscarConf(unittest.TestCase):
    '''
    vasp/poscar format named ["POSCAR_1","POSCAR_2","POSCAR_3"].
    Add optional_input["conf_format"] and optional_artifact. 
    '''
    def setUp(self):
        self.ntasks = 2
        confs = dump_conf_from_poscar("deepmd/npy",[POSCAR_1_content, POSCAR_2_content])
        self.confs = [Path(ii) for ii in confs]
        self.incar = 'incar'
        Path(self.incar).write_text('here incar')
        self.potcar = 'potcar'
        Path(self.potcar).write_text('here potcar')
        self.type_map = ['Na']
        self.optional_file = Path('optional_test')
        self.optional_file.write_text('here test')

    def tearDown(self):
        for ii in range(self.ntasks):
            work_path = Path("task.%06d"%ii)
            if work_path.is_dir():
                shutil.rmtree(work_path)
        for ii in [self.incar,self.potcar,self.optional_file]:
            if Path(ii).is_file:
                os.remove(ii)
        if not self.confs:
            for ii in self.confs:
                if ii.is_file:
                    os.remove(ii)

    def test(self):
        wf = Workflow(name = "test")
        vi = VaspInputs(0.3,self.incar,{'Na':self.potcar},True)
        vasp = Step(
            name="PrepVasp",
            template=PythonOPTemplate(PrepVasp,image=default_image),
            artifacts={
                "confs":upload_artifact(self.confs),
                "optional_artifact":upload_artifact({"TEST":Path("optional_test")}),
            },
            parameters={
                "prep_image_config" : {},
                "inputs" : vi ,
                "optional_input" : {},
                "type_map" : self.type_map,
            }
        ) 
        wf.add(vasp)
        wf.submit()

        while wf.query_status() in ["Pending","Running"]:
            time.sleep(4)
        assert(wf.query_status() == 'Succeeded')
        step = wf.query_step(name="PrepVasp")[0]
        download_artifact(step.outputs.artifacts["task_paths"])

        tdirs = check_vasp_tasks(self, self.ntasks)
        self.assertEqual(tdirs, step.outputs.parameters['task_names'].value)
        
        #check optional_artifact
        for ii in step.outputs.parameters['task_names'].value:
            self.assertEqual(Path(Path(ii)/'TEST').read_text(), "here test")

    def testWithoutOptionalParameter(self):
        wf = Workflow(name = "test")
        vi = VaspInputs(0.3,self.incar,{'Na':self.potcar},True)
        vasp = Step(
            name="PrepVasp",
            template=PythonOPTemplate(PrepVasp,image=default_image),
            artifacts={
                "confs":upload_artifact(self.confs),
                "optional_artifact":upload_artifact({"TEST":Path("optional_test")}),
            },
            parameters={
                "inputs" : vi ,
                "type_map" : self.type_map,
            }
        ) 
        wf.add(vasp)
        wf.submit()

        while wf.query_status() in ["Pending","Running"]:
            time.sleep(4)
        assert(wf.query_status() == 'Succeeded')
        step = wf.query_step(name="PrepVasp")[0]
        download_artifact(step.outputs.artifacts["task_paths"])

        tdirs = check_vasp_tasks(self, self.ntasks)
        self.assertEqual(tdirs, step.outputs.parameters['task_names'].value)
        
        #check optional_artifact
        for ii in step.outputs.parameters['task_names'].value:
            self.assertEqual(Path(Path(ii)/'TEST').read_text(), "here test")

